import subprocess
from importlib import resources
from pathlib import Path
from typing import Union

import pandas as pd


def apply_STdeconvolve(NN_dir: Union[str, Path],
                       exp_df: pd.DataFrame,
                       ct_num: int,
                       gen_ct_embedding: bool = False) -> pd.DataFrame:
    """
    Apply STdeconvolve to spot-level data.
    :param NN_dir: Niche network directory.
    :param pd.DataFram: Expression dataFrame.  #spot x #gene
    :param ct_num: Number of cell types.
    :param gen_ct_embedding: Generate cell type embedding or not.
    :return: Deconvoluted expression DataFrame.  #spot x #cell_type
    """

    # save expression matrix in csv format
    exp_matrix_file = f'{NN_dir}/filtered_spot_exp.csv.gz'
    exp_df.to_csv(exp_matrix_file, index=True, header=True, compression='gzip')

    spot_x_cell_type_file = 'spot_x_celltype_deconvolution.csv'
    ct_x_gene_file = 'celltype_x_gene_deconvolution.csv'

    with resources.path("ONTraC.external", "STdeconvolve.R") as stdeconvolve_script_path:
        try:
            stdeconvolve_script_path_str = str(stdeconvolve_script_path)
            if gen_ct_embedding:
                subprocess.run(
                    f'Rscript {stdeconvolve_script_path_str} {exp_matrix_file} {ct_num} {str(NN_dir)} {spot_x_cell_type_file} {ct_x_gene_file}',
                    shell=True)
            else:
                subprocess.run(
                    f'Rscript {stdeconvolve_script_path_str} {exp_matrix_file} {ct_num} {str(NN_dir)} {spot_x_cell_type_file}',
                    shell=True)
        except subprocess.CalledProcessError as e:
            print("Error in running R script:", e)
            print("R script stderr:", e.stderr)

    # handle deconvoluted results
    #1) The first column name missing STdeconvolve outputed csv file
    #2) "-" in spot_id will be converted to "."
    ## spot_x_cell_type_file
    spot_x_cell_type_df = pd.read_csv(Path(NN_dir).joinpath(spot_x_cell_type_file))
    spot_x_cell_type_df.index = exp_df.columns
    spot_x_cell_type_df.to_csv(Path(NN_dir).joinpath(spot_x_cell_type_file),
                               index=True,
                               header=True,
                               index_label='Spot_ID')
    ## ct_x_gene_file
    ct_x_gene_df = pd.read_csv(Path(NN_dir).joinpath(ct_x_gene_file))
    ct_x_gene_df.to_csv(Path(NN_dir).joinpath(ct_x_gene_file), index=True, header=True, index_label='Cell_Type')

    # load deconvoluted cell type composition data
    ct_coding_df = pd.read_csv(Path(NN_dir).joinpath(spot_x_cell_type_file), index_col=0)

    return ct_coding_df



from typing import List

import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix

from ..log import *


def perform_pca(expression_data: pd.DataFrame, n_components: int = 50) -> np.ndarray:
    """
    Perform PCA on the expression data.
    :param expression_data: Expression data.
    :param n_components: Number of components.
    :return: PCA result.
    """

    from sklearn.decomposition import PCA

    maximum_components = min(expression_data.shape[0], expression_data.shape[1])
    if n_components > maximum_components:
        n_components = maximum_components - 1

    model = PCA(n_components=n_components, svd_solver='arpack', random_state=0)
    info(f'Performing PCA with {n_components} components...')

    return model.fit_transform(X=expression_data.values)


def perform_harmony(embedding, meta_df: pd.DataFrame, batch_key: str) -> np.ndarray:
    """
    Perform Harmony on the embedding.
    :param embedding: Embedding.
    :param meta_df: Meta data.
    :param batch_key: Batch key.
    :return: Harmony result.
    """

    from harmonypy import run_harmony

    info(f'Performing Harmony using "{batch_key}" as batch ...')

    return run_harmony(data_mat=embedding, meta_data=meta_df, vars_use=batch_key).Z_corr.T


def define_neighbors(embedding: np.ndarray, n_neighbors: int = 20) -> csr_matrix:
    """
    Define neighbors.
    :param embedding: Embedding.
    :param n_neighbors: Number of neighbors.
    :return: Neighbors.
    """

    #TODO: n_neighbors should be less than the number of cells in each sample.

    import pynndescent
    from scipy.sparse import coo_matrix
    from umap.umap_ import fuzzy_simplicial_set

    info(f'Defining neighbors with {n_neighbors} neighbors...')

    index = pynndescent.NNDescent(embedding, n_neighbors=n_neighbors, metric='euclidean', n_jobs=4)
    knn_indices, knn_dists = index.neighbor_graph
    connectivities, _, _, _ = fuzzy_simplicial_set(X=coo_matrix(([], ([], [])), shape=(embedding.shape[0], 1)),
                                                   n_neighbors=n_neighbors,
                                                   random_state=0,
                                                   metric="euclidean",
                                                   knn_indices=knn_indices,
                                                   knn_dists=knn_dists,
                                                   return_dists=False)

    return connectivities


def perform_umap(embedding: np.ndarray,
                 n_neighbors: int = 20,
                 min_dist: float = 0.5,
                 n_components: int = 2) -> np.ndarray:
    """
    Perform UMAP on the embedding.
    :param embedding: Embedding.
    :param n_neighbors: Number of neighbors.
    :param min_dist: Minimum distance.
    :param n_components: Number of components.
    :return: UMAP result.
    """

    import umap

    info(f'Performing UMAP with {n_neighbors} neighbors, {min_dist} minimum distance, and {n_components} components...')

    return umap.UMAP(n_neighbors=n_neighbors, min_dist=min_dist, n_components=n_components).fit_transform(embedding)


def perform_leiden(connectivities: csr_matrix, resolution: float = 1.0) -> List[int]:
    """
    Perform Leiden algorithm.
    :param connectivities: Connectivities.
    :param resolution: Resolution.
    :return: Leiden result.
    """

    import igraph as ig
    import leidenalg

    sources, targets = connectivities.nonzero()
    g = ig.Graph(directed=None)
    g.add_vertices(connectivities.shape[0])
    g.add_edges(es=list(zip(sources, targets)), attributes={'weight': connectivities[sources, targets].tolist()[0]})

    info(f'Performing Leiden algorithm with resolution {resolution}...')

    partition = leidenalg.find_partition(graph=g,
                                         partition_type=leidenalg.RBConfigurationVertexPartition,
                                         resolution_parameter=resolution)

    return partition.membership


from .version import __version__

citation = """Wang, W.*, Zheng, S.*, Shin, C. S., Chávez-Fuentes J. C. & Yuan, G.-C$. ONTraC characterizes spatially continuous variations of tissue microenvironment through niche trajectory analysis. https://doi.org/10.1186/s13059-025-03588-5. Genome Biol 26: 117 (2025).

If you are using kernel-based cell type adjustment or working on low-resolution data, please also cite: Wang, W., Shin, C. S., Chávez-Fuentes J. C.  & Yuan, G.-C.$. A Robust Kernel-Based Workflow for Niche Trajectory Analysis. https://onlinelibrary.wiley.com/doi/10.1002/smtd.202401199. Small Methods, 2025."""



import os
from pathlib import Path
from typing import Dict, Optional, Tuple, Union

import numpy as np
import pandas as pd
from numpy import ndarray
from pandas import DataFrame
from torch_geometric.loader import DenseDataLoader

from ..data import SpatailOmicsDataset, load_dataset
from ..external.deconvolution import apply_STdeconvolve
from ..log import error, info
from ..utils import get_meta_data_file
from .data import load_meta_data, save_cell_type_code, save_meta_data
from .expression import define_neighbors, perform_harmony, perform_leiden, perform_pca, perform_umap


def load_input_data(
    meta_input: Union[str, Path],
    NN_dir: Union[str, Path],
    exp_input: Optional[Union[str, Path]] = None,
    embedding_input: Optional[Union[str, Path]] = None,
    low_res_exp_input: Optional[Union[str, Path]] = None,
    deconvoluted_ct_composition: Optional[Union[str, Path]] = None,
    deconvoluted_exp_input: Optional[Union[str, Path]] = None,
) -> Dict[str, pd.DataFrame]:
    """
    Load data from original inputs.
    :param meta_input: str or Path, meta data file.
    :param NN_dir: str or Path, save directory.
    :return: Dict[str, pd.DataFrame], loaded data.
    """

    output = {}

    # default values
    output['embedding_data'] = None

    # ----- load meta_data -----
    meta_data_df = load_meta_data(save_dir=NN_dir, meta_data_file=meta_input)
    ids = meta_data_df.iloc[:, 0].values.tolist()
    output['meta_data'] = meta_data_df  # N x X

    # cell or spot level data?
    id_name = meta_data_df.columns[0]
    if id_name == 'Cell_ID':
        if embedding_input is not None:
            embedding_df = pd.read_csv(embedding_input, index_col=0)
            # TODO: error message module
            if embedding_df.index.tolist() != ids:
                raise ValueError('The first column of embedding input should be same as the first column of meta_data.')
            output['embedding_data'] = embedding_df  # N x #embedding
        elif exp_input is not None:
            exp_df = pd.read_csv(exp_input, index_col=0)
            # TODO: error message module
            if exp_df.index.tolist() != ids:
                raise ValueError('The first column of exp input should be same as the first column of meta_data.')
            output['exp_data'] = exp_df  # N x #gene
    else:
        if low_res_exp_input is not None:
            low_res_exp_df = pd.read_csv(low_res_exp_input, index_col=0)
            # TODO: error message module
            if low_res_exp_df.columns.tolist() != ids:
                raise ValueError('The first row in low_res_exp_data should be same as the first column of meta_data.')
            output['low_res_exp'] = low_res_exp_df  # #gene x #N
        if deconvoluted_ct_composition is not None:
            deconvoluted_ct_composition_df = pd.read_csv(deconvoluted_ct_composition, index_col=0)
            deconvoluted_ct_composition_df.index = deconvoluted_ct_composition_df.index.astype(str)
            deconvoluted_ct_composition_df.columns = deconvoluted_ct_composition_df.columns.astype(str)
            # TODO: error message module
            if deconvoluted_ct_composition_df.shape[0] != meta_data_df.shape[0]:
                raise ValueError('The number of rows in meta_data and deconvoluted_ct_composition should be the same.')
            output['deconvoluted_ct_composition'] = deconvoluted_ct_composition_df  # N x #cell_type
        if deconvoluted_exp_input is not None and deconvoluted_ct_composition is not None:  # if deconvoluted_exp_input is provided, deconvoluted_ct_composition should also be provided
            deconvoluted_exp_df = pd.read_csv(deconvoluted_exp_input, index_col=0)
            deconvoluted_exp_df.index = deconvoluted_exp_df.index.astype(str)
            deconvoluted_exp_df.columns = deconvoluted_exp_df.columns.astype(str)
            # TODO: error message module
            if deconvoluted_exp_df.index.tolist() != deconvoluted_ct_composition_df.columns.tolist():
                raise ValueError(
                    'The index of deconvoluted_exp_data should be same as the columns of deconvoluted_ct_composition.')
            output['deconvoluted_exp_data'] = deconvoluted_exp_df  # #cell_type x #gene

    return output


def perform_deconvolution(NN_dir: Union[str, Path],
                          dc_method: str,
                          exp_df: pd.DataFrame,
                          dc_ct_num: int,
                          gen_ct_embedding: bool = False) -> pd.DataFrame:
    """
    Perform deconvolution.
    :param NN_dir: str or Path, save directory.
    :param dc_method: str, deconvolution method.
    :param exp_df: pd.DataFrame, expression matrix.  #gene x #spot
    :param dc_ct_num: int, number of cell types.
    :param gen_ct_embedding: Generate cell type embedding or not.
    :return: np.ndarray, deconvoluted cell type matrix.  #spot x #cell_type
    """

    info(message='            -------- deconvolution -------           ')

    if dc_method == 'STdeconvolve':
        info(message='Apply STdeconvolve to low resolution data.')
        ct_coding_df = apply_STdeconvolve(NN_dir=NN_dir,
                                          exp_df=exp_df,
                                          ct_num=dc_ct_num,
                                          gen_ct_embedding=gen_ct_embedding)

    return ct_coding_df


def cal_cell_type_coding(
    input_data: Dict[str, pd.DataFrame],
    NN_dir: Union[str, Path],
    resolution: Optional[float] = None,
    dc_method: Optional[str] = None,
    dc_ct_num: Optional[int] = None,
    gen_ct_embedding: bool = False,
) -> Dict[str, pd.DataFrame]:
    """
    """

    meta_data_df = input_data['meta_data']

    id_name = meta_data_df.columns[0]
    if id_name == 'Cell_ID':
        if 'Cell_Type' in meta_data_df.columns:  # option 1: cell-level data with Cell_Type info in meta_data
            meta_data_df['Cell_Type'] = meta_data_df['Cell_Type'].astype('category')
            if gen_ct_embedding and input_data.get('embedding_data', None) is None:  # generate cell type embedding
                if input_data.get('exp_data', None) is None:
                    raise ValueError(
                        'exp_data or embedding_data should be provided for cell-level data when generating cell type embedding.'
                    )
                else:
                    pca_embedding = perform_pca(input_data['exp_data'])
                    if 'Batch' in meta_data_df.columns and meta_data_df['Batch'].nunique() > 1:
                        pca_embedding = perform_harmony(pca_embedding, meta_data_df, 'Batch')
                # save PCA embedding
                input_data['embedding_data'] = pd.DataFrame(data=pca_embedding,
                                                            columns=[f'PC{i+1}' for i in range(pca_embedding.shape[1])],
                                                            index=meta_data_df[id_name])
                np.savetxt(Path(NN_dir).joinpath('PCA_embedding.csv'), pca_embedding, delimiter=',')

        else:
            if input_data.get('embedding_data', None) is not None:  # option 2: cell-level data with embedding info
                pca_embedding = input_data['embedding_data'].values
            elif input_data.get('exp_data', None) is not None:  # option 3: cell-level data with gene expression data
                pca_embedding = perform_pca(input_data['exp_data'])
                if 'Batch' in meta_data_df.columns and meta_data_df['Batch'].nunique() > 1:
                    pca_embedding = perform_harmony(pca_embedding, meta_data_df, 'Batch')
            else:
                raise ValueError(
                    'Cell_Type column, exp_data, or embedding_data should be provided for cell-level data.')

            # save PCA embedding
            input_data['embedding_data'] = pd.DataFrame(data=pca_embedding,
                                                        columns=[f'PC{i+1}' for i in range(pca_embedding.shape[1])],
                                                        index=meta_data_df[id_name])
            np.savetxt(Path(NN_dir).joinpath('PCA_embedding.csv'), pca_embedding, delimiter=',')

            if resolution is None:
                raise ValueError(
                    'resolution should be provided for cell-level data with gene expression or embedding data as input when Cell_Type is not provided in meta data input.'
                )

            connectivities = define_neighbors(pca_embedding)
            leiden_result = perform_leiden(connectivities, resolution=resolution)
            umap_embedding = perform_umap(pca_embedding)
            np.savetxt(Path(NN_dir).joinpath('UMAP_embedding.csv'), umap_embedding, delimiter=',')
            meta_data_df['Cell_Type'] = pd.Categorical(leiden_result)
            input_data['meta_data'] = meta_data_df
            save_meta_data(save_dir=NN_dir, meta_data_df=meta_data_df)

            # save meta data
            meta_data_df.to_csv(Path(NN_dir).joinpath('meta_data.csv'), index=False)

        # generate cell type coding matrix
        ct_coding_matrix = np.zeros(shape=(meta_data_df.shape[0],
                                           meta_data_df['Cell_Type'].cat.categories.shape[0]))  # N x #cell_type
        ct_coding_matrix[np.arange(meta_data_df.shape[0]), meta_data_df.Cell_Type.cat.codes.values] = 1
        ct_coding_df = pd.DataFrame(data=ct_coding_matrix,
                                    columns=meta_data_df['Cell_Type'].cat.categories,
                                    index=meta_data_df[id_name])
        # save cell type code
        save_cell_type_code(save_dir=NN_dir, cell_types=meta_data_df['Cell_Type'])

    else:  # id_name == 'Spot_ID', low resolution data
        if 'deconvoluted_ct_composition' in input_data:  # option 4: low resolution data with deconvoluted cell type composition
            ct_coding_df = input_data['deconvoluted_ct_composition']  # N x #cell_type
        elif 'low_res_exp' in input_data:  # option 5: low resolution data with original expression data
            if dc_method is None or dc_ct_num is None:
                raise ValueError('dc_method and dc_ct_num are required when you provide low_res_exp_data as input.')

            ct_coding_df = perform_deconvolution(NN_dir=NN_dir,
                                                 dc_method=dc_method,
                                                 exp_df=input_data['low_res_exp'],
                                                 dc_ct_num=dc_ct_num,
                                                 gen_ct_embedding=gen_ct_embedding)
        else:
            raise ValueError(
                'deconvoluted_ct_composition or low_res_exp_data should be provided for low resolution data.')

        # save cell type code
        save_cell_type_code(save_dir=NN_dir, cell_types=pd.Series(ct_coding_df.columns))

    input_data['ct_coding'] = ct_coding_df
    ct_coding_df.to_csv(Path(NN_dir).joinpath('ct_coding.csv'), index=True, index_label=id_name)

    return input_data


def preprocessing_nn(meta_input: Union[str, Path],
                     NN_dir: Union[str, Path],
                     exp_input: Optional[Union[str, Path]] = None,
                     embedding_input: Optional[Union[str, Path]] = None,
                     low_res_exp_input: Optional[Union[str, Path]] = None,
                     gen_ct_embedding: bool = False,
                     deconvoluted_ct_composition: Optional[Union[str, Path]] = None,
                     deconvoluted_exp_input: Optional[Union[str, Path]] = None,
                     resolution: Optional[float] = None,
                     dc_method: Optional[str] = None,
                     dc_ct_num: Optional[int] = None) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Preprocessing for niche network.

    Possible input parameters combinations:
    # class I: without embedding ajustment
    1. meta_input, NN_dir (cell-level data with Cell_Type info in meta_data)
    2. meta_input, NN_dir, exp_input, resolution (cell-level data with gene expression data)
    3. meta_input, NN_dir, embedding_input (cell-level data with embedding info)
    4. meta_input, NN_dir, low_res_exp_input, dc_method, dc_ct_num (spot-level data with original expression data)
    5. meta_input, NN_dir, low_res_exp_input, deconvoluted_ct_composition (spot-level data with deconvoluted cell type composition)
    # class II: with embedding ajustment
    1. embedding_adjust, sigma, meta_input, NN_dir, exp_input, resolution  (cell-level data with gene expression data)
    2. embedding_adjust, sigma, meta_input, NN_dir, embedding_input (cell-level data with embedding info)
    3. meta_input, NN_dir, low_res_exp_input, dc_method, dc_ct_num (spot-level data with original expression data)
    4. meta_input, NN_dir, low_res_exp_input, deconvoluted_ct_composition, deconvoluted_exp_input (spot-level data with deconvoluted cell type composition)

    Steps:
    1. define cell type
    2. cell type coding matrix
    3. embedding adjustment (optional)


    :param meta_input: str or Path, meta data file.
    :param NN_dir: str or Path, save directory.
    :param exp_input: str or Path, expression data file.
    :param embedding_input: str or Path, embedding data file.
    :param low_res_exp_input: str or Path, low resolution expression data file.
    :param gen_ct_embedding: bool, generate cell type embedding.
    :param deconvoluted_ct_composition: str or Path, deconvoluted cell type composition file.
    :param deconvoluted_exp_input: str or Path, deconvoluted expression data file.
    :param resolution: float, resolution.
    :param dc_method: str, deconvolution method.
    :param dc_ct_num: int, deconvoluted cell type number.
    :param embedding_adjust: bool, adjust embedding.
    :param sigma: float, sigma.
    :return: Tuple[pd.DataFrame, pd.DataFrame], meta data and cell type coding matrix.
    """

    input_data = load_input_data(meta_input=meta_input,
                                 NN_dir=NN_dir,
                                 exp_input=exp_input,
                                 embedding_input=embedding_input,
                                 low_res_exp_input=low_res_exp_input,
                                 deconvoluted_ct_composition=deconvoluted_ct_composition,
                                 deconvoluted_exp_input=deconvoluted_exp_input)

    input_data = cal_cell_type_coding(input_data=input_data,
                                      NN_dir=NN_dir,
                                      resolution=resolution,
                                      dc_method=dc_method,
                                      dc_ct_num=dc_ct_num,
                                      gen_ct_embedding=gen_ct_embedding)

    return input_data['meta_data'], input_data['embedding_data'], input_data['ct_coding']


def load_data(NN_dir: Union[str, Path], batch_size: int = 0) -> Tuple[SpatailOmicsDataset, DenseDataLoader]:
    """
    Load data and create sample loader.
    :param NN_dir: str or Path, save directory.
    :param batch_size: int, batch size.
    :return: Tuple[SpatailOmicsDataset, DenseDataLoader], dataset and sample loader.
    """

    info('Loading dataset.')

    dataset = load_dataset(NN_dir=NN_dir)
    batch_size = batch_size if batch_size > 0 else len(dataset)
    sample_loader = DenseDataLoader(dataset, batch_size=batch_size)

    return dataset, sample_loader


def preprocessing_gnn(NN_dir: Union[str, Path],
                      batch_size: int = 0) -> Tuple[SpatailOmicsDataset, DenseDataLoader, pd.DataFrame]:
    """
    Preprocessing for GNN.
    :param NN_dir: str or Path, save directory.
    :param batch_size: int, batch size.
    :return: Tuple[SpatailOmicsDataset, DenseDataLoader, pd.DataFrame], dataset, sample loader, and meta data.
    """

    # meta data
    meta_data_df = pd.read_csv(get_meta_data_file(NN_dir), header=0)
    meta_data_df['Sample'] = meta_data_df['Sample'].astype(str).astype('category')
    if 'Cell_Type' in meta_data_df.columns:
        meta_data_df['Cell_Type'] = meta_data_df['Cell_Type'].astype(str).astype('category')

    # dataset and sample loader
    dataset, sample_loader = load_data(NN_dir=NN_dir, batch_size=batch_size)

    return dataset, sample_loader, meta_data_df


def preprocessing_nt(NN_dir: Union[str, Path], GNN_dir: Union[str, Path]) -> Tuple[DataFrame, DataFrame, ndarray]:
    """
    Preprocessing for niche trajectory.
    :param NN_dir: str or Path, save directory.
    :param GNN_dir: str or Path, save directory.
    :return: Tuple[DataFrame, DataFrame, ndarray], meta data, niche-level niche cluster assign, and consolidate out_adj.
    """

    # params
    niche_level_niche_cluster_file = Path(f'{GNN_dir}/niche_level_niche_cluster.csv.gz')
    consolidate_out_adj_file = Path(f'{GNN_dir}/consolidate_out_adj.csv.gz')

    # load meta data
    meta_data_df = pd.read_csv(get_meta_data_file(NN_dir), header=0)
    meta_data_df['Sample'] = meta_data_df['Sample'].astype(str).astype('category')
    meta_data_df['Cell_Type'] = meta_data_df['Cell_Type'].astype(str).astype('category')

    # load niche-level niche cluster assign
    if not os.path.exists(niche_level_niche_cluster_file) or not os.path.exists(consolidate_out_adj_file):
        error(f'niche_level_niche_cluster.csv.gz or consolidate_out_adj.csv.gz does not exist in {GNN_dir} directory.')

    niche_level_niche_cluster_assign_df = pd.read_csv(niche_level_niche_cluster_file, header=0, index_col=0)

    # load consolidate out_adj
    consolidate_out_adj_array = np.loadtxt(consolidate_out_adj_file, delimiter=',')

    return meta_data_df, niche_level_niche_cluster_assign_df, consolidate_out_adj_array



# pre-processing options

import sys
from optparse import OptionGroup, OptionParser, Values
from typing import Optional

from ..log import *
from ..version import __version__
from ._IO import *


# ------------------------------------
# Functions
# ------------------------------------
def add_preprocessing_options_group(optparser: OptionParser) -> None:
    """
    Add preprocessing options group to optparser.
    :param optparser: OptionParser object.
    :param preprocessing_options: Set of preprocessing options.
    :return: OptionGroup object.
    """

    # preprocessing options group
    group_preprocessing = OptionGroup(optparser, "Preprocessing")
    group_preprocessing.add_option(
        '--resolution',
        dest='resolution',
        type=float,
        default=10.0,
        help=
        'Resolution for leiden clustering. Used for clustering cells into cell types when gene expression data is provided. Default is 10.0.'
    )
    group_preprocessing.add_option('--deconvolution-method',
                                   dest='dc_method',
                                   default='STdeconvolve',
                                   choices=['STdeconvolve'],
                                   help='Deconvolution method used for low resolution data. Default is STdeconvolve.')
    group_preprocessing.add_option('--deconvolution-ct-num',
                                   dest='dc_ct_num',
                                   type='int',
                                   default=10,
                                   help='Number of cell type that the deconvolution method will deconvolve.')
    optparser.add_option_group(group_preprocessing)


def validate_preprocessing_options(options: Values, optparser: Optional[OptionParser] = None) -> None:
    """
    Validate preprocessing options.
    :param options: Options object.
    :param optparser: OptionParser object.
    :return: None
    """

    if options.exp_input is None and options.embedding_input is None:  # no cell-level expression data or embedding data
        info(
            message=
            'No expression data or embedding data is provided. ONTraC will not perform any preprocessing for cell-level data.'
        )
        options.resolution = None
    elif options.resolution < 0:
        error('resolution must be greater than 0.')
        if optparser is not None: optparser.print_help()
        sys.exit(1)

    if options.low_res_exp_input is None:  # low resolution data is not provided
        options.dc_method = None
        options.dc_ct_num = None
    else:  # low resolution data is provided
        if options.dc_ct_num < 2:
            error('deconvolution_cell_type_number must be greater than 2.')
            if optparser is not None: optparser.print_help()
            sys.exit(1)
        if options.dc_ct_num < 4:
            warning('deconvolution_cell_type_number is less than 4. The result may not be reliable.')


def write_preprocessing_memo(options: Values):
    """Write preprocessing memos to stdout.

    Args:
        options: Options object.
    """

    # print parameters to stdout
    info(message='      -------- preprocessing options -------      ')
    if options.resolution is not None:
        info(message=f'resolution: {options.resolution}')
    if options.dc_method is not None:
        info(message=f'deconvolution method: {options.dc_method}')
        info(message=f'deconvolution cell type number: {options.dc_ct_num}')



from optparse import Values

from ..analysis.cell_type import cell_type_visualization
from ..analysis.data import AnaData
from ..analysis.niche_cluster import niche_cluster_visualization
from ..analysis.niche_net import clustering_visualization_from_anadata, embedding_adjust_visualization_from_anadata
from ..analysis.spatial import spatial_visualization
from ..analysis.train_loss import train_loss_visualiztion
from ..log import *
from ..optparser import opt_analysis_validate, prepare_analysis_optparser
from ..run.processes import load_parameters
from ..utils import *
from ..version import __version__


# ------------------------------------
# Functions
# ------------------------------------
def analysis_pipeline(options: Values) -> None:
    """
    ONTraC analysis pipeline
    """

    info('-------------- Analysis pipeline start -------------- ')
    step_index = 0

    # 0. load data class
    info(message=f'Analysis pipeline step {step_index}: load data.')
    ana_data = AnaData(options)
    step_index += 1

    step_index = 0
    if ana_data.options.spatial_res == 'cell':
        # part 1: clustering
        info(message=f'Analysis pipeline step {step_index}: clustering visualization.')
        clustering_visualization_from_anadata(ana_data=ana_data)
        step_index += 1

    if ana_data.options.embedding_adjust:
        # part 2: niche network construction
        info(message=f'Analysis pipeline step {step_index}: embedding adjust visualization.')
        embedding_adjust_visualization_from_anadata(ana_data=ana_data)
        step_index += 1

    # part 3: train loss
    info(message=f'Analysis pipeline step {step_index}: train loss visualization.')
    train_loss_visualiztion(ana_data=ana_data)
    step_index += 1

    # part 4: spatial based output
    info(message=f'Analysis pipeline step {step_index}: spatial-based visualization.')
    spatial_visualization(ana_data=ana_data)
    step_index += 1

    # part 5: niche cluster
    info(message=f'Analysis pipeline step {step_index}: niche cluster visualization.')
    niche_cluster_visualization(ana_data=ana_data)
    step_index += 1

    # part 6: cell type based output
    info(message=f'Analysis pipeline step {step_index}: cell type-based visualization.')
    cell_type_visualization(ana_data=ana_data)
    step_index += 1
    info('--------------- Analysis pipeline end --------------- ')


def main():
    """
    The main function
    """

    # write version information
    write_version_info()

    options = load_parameters(opt_validate_func=opt_analysis_validate,
                              prepare_optparser_func=prepare_analysis_optparser)

    analysis_pipeline(options)


if __name__ == '__main__':
    main()



from pathlib import Path
from typing import List, Optional, Tuple, Union

import matplotlib as mpl
import numpy as np
import pandas as pd
from scipy.spatial import distance
from seaborn.matrix import ClusterGrid

mpl.rcParams['pdf.fonttype'] = 42
mpl.rcParams['ps.fonttype'] = 42
mpl.rcParams['font.family'] = 'Arial'
import matplotlib.pyplot as plt
import seaborn as sns

from ..log import info, warning
from .data import AnaData


def clustering_visualization(
        data_df: pd.DataFrame,
        output_file_path: Optional[Union[str, Path]] = None) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """
    Visualization of clustering results.
    :param data_df: pd.DataFrame, data for visualization.
    :param output_file_path: str or Path, output file path.
    :return: None.
    """

    with sns.axes_style('white', rc={
            'xtick.bottom': True,
            'ytick.left': True
    }), sns.plotting_context('paper',
                             rc={
                                 'axes.titlesize': 8,
                                 'axes.labelsize': 8,
                                 'xtick.labelsize': 6,
                                 'ytick.labelsize': 6,
                                 'legend.fontsize': 6
                             }):
        fig, ax = plt.subplots(1, 1, figsize=(8, 5))
        sns.scatterplot(data=data_df, x='Embedding_1', y='Embedding_2', hue='Cell_Type', s=2, ax=ax)
        ax.set_xlabel('UMAP_1')
        ax.set_ylabel('UMAP_2')
        ax.legend(loc='upper left', bbox_to_anchor=(1, 1), ncol=3, markerscale=4)
        fig.tight_layout()
        if output_file_path:
            fig.savefig(f'{output_file_path}/clustering.pdf', transparent=True)
            return None
        else:
            return fig


def clustering_visualization_from_anadata(ana_data: AnaData) -> Optional[Tuple[plt.Figure, plt.Axes]]:
    """Visualization of clustering results.
    
    Args:
        ana_data: AnaData object.
    """

    # check if the embedding is available
    if ana_data.umap_embedding is None:
        warning('UMAP embedding is not available. Skip the clustering visualization.')
        return None

    data_df = pd.DataFrame(ana_data.umap_embedding, columns=['Embedding_1', 'Embedding_2'])
    data_df.index = ana_data.meta_data_df.index
    data_df['Cell_Type'] = ana_data.meta_data_df['Cell_Type']
    data_df['Cell_Type'] = data_df['Cell_Type'].astype('category')

    return clustering_visualization(data_df=data_df, output_file_path=ana_data.options.output)


def embedding_adjust_visualization(dis_df: pd.DataFrame,
                                   output_file_name: str,
                                   output_file_path: Optional[Union[str, Path]] = None) -> Optional[ClusterGrid]:
    with sns.axes_style('white', rc={
            'xtick.bottom': True,
            'ytick.left': True
    }), sns.plotting_context('paper',
                             rc={
                                 'axes.titlesize': 8,
                                 'axes.labelsize': 8,
                                 'xtick.labelsize': 6,
                                 'ytick.labelsize': 6,
                                 'legend.fontsize': 6
                             }):
        cluster_grid: ClusterGrid = sns.clustermap(dis_df, figsize=(dis_df.shape[0] / 6, dis_df.shape[0] / 6))
        if output_file_path is not None:
            cluster_grid.savefig(f'{output_file_path}/{output_file_name}', transparent=True)
            return None
        else:
            return cluster_grid


def embedding_adjust_visualization_from_anadata(ana_data: AnaData) -> Optional[List[ClusterGrid]]:
    """Visualization of embedding adjust.

    Args:
        ana_data: AnaData object.
    """

    if not ana_data.options.embedding_adjust:
        return None

    if ana_data.ct_embedding is None:
        warning('Cell type embedding is not available. Skip the embedding adjustment visualization.')
        return None

    cell_types = ana_data.cell_type_codes['Cell_Type'].tolist()
    raw_distance = distance.cdist(ana_data.ct_embedding.values, ana_data.ct_embedding.values, 'euclidean')

    # calculate distance between each cell type
    median_distance = np.median(raw_distance[np.triu_indices(raw_distance.shape[0], k=1)])
    info(f'Median distance between cell types: {median_distance}')
    raw_distance_df = pd.DataFrame(raw_distance, index=cell_types, columns=cell_types)

    if ana_data.options.sigma is None:
        ana_data.options.sigma = np.median(raw_distance[np.triu_indices(raw_distance.shape[0], k=1)])
        info(
            f'Sigma is not provided. Use the median ({ana_data.options.sigma}) of the distances between the cell type pairs.'
        )

    # calculate the M
    M = np.exp(-(raw_distance / (ana_data.options.sigma * median_distance))**2)
    M_df = pd.DataFrame(M, index=cell_types, columns=cell_types)

    output = []
    output.append(
        embedding_adjust_visualization(dis_df=raw_distance_df,
                                       output_file_name='raw_distance.pdf',
                                       output_file_path=ana_data.options.output))
    output.append(
        embedding_adjust_visualization(dis_df=M_df,
                                       output_file_name='adjusted_distance.pdf',
                                       output_file_path=ana_data.options.output))
    return output
